@@ -4,15 +4,17 @@
 //! RTP and RTCP handling; see [RFC 3550](https://datatracker.ietf.org/doc/html/rfc3550).
 
 use bytes::{Buf, Bytes};
-use log::debug;
+use log::{debug, trace};
 use pretty_hex::PrettyHex;
 
 use crate::client::PacketItem;
-use crate::{Error, ErrorInt};
+use crate::{ConnectionContext, Error, ErrorInt, PacketContext};
+
+use super::{SessionOptions, Timeline};
 
 /// A received RTP packet.
 pub struct Packet {
-    pub ctx: crate::PacketContext,
+    pub ctx: PacketContext,
     pub stream_id: usize,
     pub timestamp: crate::Timestamp,
     pub ssrc: u32,
@@ -51,7 +53,7 @@ impl std::fmt::Debug for Packet {
 #[derive(Debug)]
 pub struct SenderReport {
     pub stream_id: usize,
-    pub ctx: crate::PacketContext,
+    pub ctx: PacketContext,
     pub timestamp: crate::Timestamp,
     pub ntp_timestamp: crate::NtpTimestamp,
 }
@@ -88,10 +90,10 @@ impl InorderParser {
 
     pub fn rtp(
         &mut self,
-        session_options: &super::SessionOptions,
-        conn_ctx: &crate::ConnectionContext,
-        pkt_ctx: &crate::PacketContext,
-        timeline: &mut super::Timeline,
+        session_options: &SessionOptions,
+        conn_ctx: &ConnectionContext,
+        pkt_ctx: &PacketContext,
+        timeline: &mut Timeline,
         stream_id: usize,
         mut data: Bytes,
     ) -> Result<Option<PacketItem>, Error> {
@@ -108,10 +110,26 @@ impl InorderParser {
                 ),
             })
         })?;
+
+        // Skip pt=50 packets, sent by at least Geovision cameras. I'm not sure
+        // what purpose these serve, but they have the same sequence number as
+        // the packet immediately before. In TCP streams, this can cause an
+        // "Out-of-order packet or large loss" error. In UDP streams, if these
+        // are delivered out of order, they will cause the more important other
+        // packet with the same sequence number to be skipped.
+        if reader.payload_type() == 50 {
+            debug!("skipping pkt with invalid payload type 50");
+            return Ok(None);
+        }
+
         let sequence_number = u16::from_be_bytes([data[2], data[3]]); // I don't like rtsp_rs::Seq.
         let ssrc = reader.ssrc();
         let loss = sequence_number.wrapping_sub(self.next_seq.unwrap_or(sequence_number));
+        let is_tcp = matches!(session_options.transport, super::Transport::Tcp);
         if matches!(self.ssrc, Some(s) if s != ssrc) {
+            if let (Some(session_group), true) = (session_options.session_group.as_ref(), is_tcp) {
+                session_group.note_stale_live555_data();
+            }
             bail!(ErrorInt::RtpPacketError {
                 conn_ctx: *conn_ctx,
                 pkt_ctx: *pkt_ctx,
@@ -144,6 +162,7 @@ impl InorderParser {
                     self.ssrc,
                     self.next_seq
                 );
+                return Ok(None);
             }
         }
         let timestamp = match timeline.advance_to(reader.timestamp()) {
@@ -186,8 +205,9 @@ impl InorderParser {
 
     pub fn rtcp(
         &mut self,
-        pkt_ctx: &crate::PacketContext,
-        timeline: &mut super::Timeline,
+        session_options: &SessionOptions,
+        pkt_ctx: &PacketContext,
+        timeline: &mut Timeline,
         stream_id: usize,
         data: Bytes,
     ) -> Result<Option<PacketItem>, String> {
@@ -212,6 +232,13 @@ impl InorderParser {
 
                     let ssrc = pkt.ssrc();
                     if matches!(self.ssrc, Some(s) if s != ssrc) {
+                        use super::Transport;
+                        if let (Some(session_group), Transport::Tcp) = (
+                            session_options.session_group.as_ref(),
+                            session_options.transport,
+                        ) {
+                            session_group.note_stale_live555_data();
+                        }
                         return Err(format!(
                             "Expected ssrc={:08x?}, got RTCP SR ssrc={:08x}",
                             self.ssrc, ssrc
@@ -226,10 +253,144 @@ impl InorderParser {
                         ntp_timestamp: pkt.ntp_timestamp(),
                     });
                 }
-                crate::rtcp::Packet::Unknown(pkt) => debug!("rtcp: pt {:?}", pkt.payload_type()),
+                crate::rtcp::Packet::Unknown(pkt) => trace!("rtcp: pt {:?}", pkt.payload_type()),
             }
             i += 1;
         }
         Ok(sr.map(PacketItem::SenderReport))
     }
 }
+
+#[cfg(test)]
+mod tests {
+    use super::*;
+
+    /// Checks dropping and logging Geovision's extra payload type 50 packets.
+    /// On a GV-EBD4701 running V1.02_2021_04_08, these seem to appear after
+    /// every IDR frame, with the same sequence number as the final packet in
+    /// that frame.
+    #[test]
+    fn geovision_pt50_packet() {
+        let mut timeline = Timeline::new(None, 90_000, None).unwrap();
+        let mut parser = InorderParser::new(Some(0xd25614e), None);
+
+        // Normal packet.
+        match parser.rtp(
+            &SessionOptions::default(),
+            &ConnectionContext::dummy(),
+            &PacketContext::dummy(),
+            &mut timeline,
+            0,
+            rtp_rs::RtpPacketBuilder::new()
+                .payload_type(105)
+                .ssrc(0xd25614e)
+                .sequence(0x1234.into())
+                .timestamp(141000)
+                .marked(true)
+                .payload(b"foo")
+                .build()
+                .unwrap()
+                .into(),
+        ) {
+            Ok(Some(PacketItem::RtpPacket(_))) => {}
+            o => panic!("unexpected packet 1 result: {:#?}", o),
+        }
+
+        // Mystery pt=50 packet with same sequence number.
+        match parser.rtp(
+            &SessionOptions::default(),
+            &ConnectionContext::dummy(),
+            &PacketContext::dummy(),
+            &mut timeline,
+            0,
+            rtp_rs::RtpPacketBuilder::new()
+                .payload_type(50)
+                .ssrc(0xd25614e)
+                .sequence(0x1234.into())
+                .timestamp(141000)
+                .marked(true)
+                .payload(b"bar")
+                .build()
+                .unwrap()
+                .into(),
+        ) {
+            Ok(None) => {}
+            o => panic!("unexpected packet 2 result: {:#?}", o),
+        }
+    }
+
+    #[test]
+    fn out_of_order() {
+        let mut timeline = Timeline::new(None, 90_000, None).unwrap();
+        let mut parser = InorderParser::new(Some(0xd25614e), None);
+
+        let session_options = SessionOptions::default().transport(crate::client::Transport::Udp);
+        match parser.rtp(
+            &session_options,
+            &ConnectionContext::dummy(),
+            &PacketContext::dummy(),
+            &mut timeline,
+            0,
+            rtp_rs::RtpPacketBuilder::new()
+                .payload_type(96)
+                .ssrc(0xd25614e)
+                .sequence(2.into())
+                .timestamp(2)
+                .marked(true)
+                .payload(b"pkt 2")
+                .build()
+                .unwrap()
+                .into(),
+        ) {
+            Ok(Some(PacketItem::RtpPacket(p))) => {
+                assert_eq!(p.timestamp.elapsed(), 0);
+            }
+            o => panic!("unexpected packet 2 result: {:#?}", o),
+        }
+
+        match parser.rtp(
+            &session_options,
+            &ConnectionContext::dummy(),
+            &PacketContext::dummy(),
+            &mut timeline,
+            0,
+            rtp_rs::RtpPacketBuilder::new()
+                .payload_type(96)
+                .ssrc(0xd25614e)
+                .sequence(1.into())
+                .timestamp(1)
+                .marked(true)
+                .payload(b"pkt 1")
+                .build()
+                .unwrap()
+                .into(),
+        ) {
+            Ok(None) => {}
+            o => panic!("unexpected packet 1 result: {:#?}", o),
+        }
+
+        match parser.rtp(
+            &session_options,
+            &ConnectionContext::dummy(),
+            &PacketContext::dummy(),
+            &mut timeline,
+            0,
+            rtp_rs::RtpPacketBuilder::new()
+                .payload_type(96)
+                .ssrc(0xd25614e)
+                .sequence(3.into())
+                .timestamp(3)
+                .marked(true)
+                .payload(b"pkt 3")
+                .build()
+                .unwrap()
+                .into(),
+        ) {
+            Ok(Some(PacketItem::RtpPacket(p))) => {
+                // The missing timestamp shouldn't have adjusted time.
+                assert_eq!(p.timestamp.elapsed(), 1);
+            }
+            o => panic!("unexpected packet 2 result: {:#?}", o),
+        }
+    }
+}
